#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
import contextlib
import os
import tempfile
from argparse import ArgumentParser, Namespace
from collections.abc import Iterator
from pathlib import Path
from typing import Final

pass
from pdm.cli.commands.base import BaseCommand
from pdm.project import Project
from pdm_pfsc.logging import (
    logger,
    setup_logger,
    traced_function,
    update_logger_from_project_ui,
)

from .executor import (
    ExecutionError,
    Executor,
    PdmExportDependenciesExecutor,
    PipAuditExecutor,
)


class AuditCommand(BaseCommand):
    # Justification: Fulfill a protocol
    name: Final[str] = "audit"  # pylint: disable=C0103
    description: str = "Runs an audit tool on the packages installed by PDM"

    def add_arguments(self, parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        parser: ArgumentParser :


        Returns
        -------

        """
        parser.add_argument(
            "auditor_arguments",
            nargs="*",
            default=[],
            help="Arguments passed to pip-audit - should be preceded by --\n"
            "The arguments to apply to maybe\n"
            "     -S/--strict\n"
            "     -l/--local\n"
            "     -s/--vulnerability-servers osv|pypi\n"
            "     -f/--format columns|json|cyclonedx-json|cyclonedx-xml|markdown\n"
            "     -o/--output FILE\n"
            "     -d/--dry-run\n"
            "     --timeout TIMEOUT (in seconds)",
            action="store",
        )

    def handle(self, project: Project, options: Namespace) -> None:
        """

        Parameters
        ----------
        project: _ProjectLike :

        options: Namespace :


        Returns
        -------

        """
        # This will not handling tracing or related parts
        # Should be evaluated at start-up time
        if hasattr(options, "verbose"):
            setup_logger(options.verbose)
        update_logger_from_project_ui(project.core.ui)

        args = tuple(options.auditor_arguments or [])
        auditor: Auditor = Auditor()
        auditor.audit(project, options.verbose or False, *args)


@contextlib.contextmanager
def _cwd(path: Path) -> Iterator[None]:
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        yield None
    finally:
        os.chdir(old_cwd)


class Auditor:
    @traced_function
    def audit(self, project: Project, verbose: bool, *args: str) -> None:
        with _cwd(project.root):
            logger.info("Auditing packages installed by PDM ...")
            repeatable = (
                project.config["plugin.audit.repeatable_audit"] or False
            )
            with tempfile.NamedTemporaryFile(
                suffix="req.txt",
                prefix="pdm_audit",
            ) as req_file:
                logger.debug("Exporting requirements.txt file from PDM")
                req_file_path: Path = Path(req_file.name).resolve()
                export: Executor = PdmExportDependenciesExecutor(
                    req_file_path, repeatable
                )
                audit: Executor = PipAuditExecutor(
                    req_file_path, verbose, repeatable, *args
                )

                try:
                    Executor.execute_chain(export, audit)
                except ExecutionError as e:
                    logger.exception(
                        "Auditing failed.",
                        exc_info=e,
                        stack_info=False,
                    )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from pdm_pfsc.logging import logger, traced_function
from pdm_pfsc.proc import CliRunnerMixin

from .updates import get_dependencies


class ExecutionError(Exception):
    def __init__(self, executor: "Executor") -> None:
        message = f"Failed to execute {executor.name}: {executor.description}"
        super().__init__(message)


class Executor(ABC):
    """"""

    @property
    @abstractmethod
    def name(self) -> str:
        """"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def description(self) -> str:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def execute(self) -> int:
        """"""
        raise NotImplementedError()

    @staticmethod
    def execute_chain(*executors: "Executor") -> None:
        """"""
        for executor in executors:
            if executor.execute() != 0:
                raise ExecutionError(executor)


class PdmExportDependenciesExecutor(Executor, CliRunnerMixin):
    def __init__(self, out_file: Path, write_hashes: bool = False) -> None:
        self.__out_file = out_file
        self.__write_hashes = write_hashes

    @property
    def name(self) -> str:
        """"""
        return "Export"

    @property
    def description(self) -> str:
        """"""
        return f"Export dependencies to {self.out_file}"

    @property
    def out_file(self) -> Path:
        """"""
        return self.__out_file

    @traced_function
    def execute(self) -> int:
        """"""
        pdm: Optional[Path] = self._which("pdm")
        if pdm is None:
            return -1

        args = [
            "-f",
            "requirements",
            "-G",
            ":all",
            "-o",
            str(self.out_file),
        ]

        if not self.__write_hashes:
            args.append("--no-hashes")

        call = ["export"]
        call.extend(args)

        exit_code, _, err = self.run(
            pdm,
            tuple(call),
        )

        if exit_code > 0:
            logger.error("Error while exporting lock file.")
            logger.warning(err)

        return exit_code


class PipAuditExecutor(Executor, CliRunnerMixin):
    """"""

    def __init__(
        self,
        input_file: Path,
        verbose: bool = False,
        repeatable: bool = False,
        *args: str,
    ) -> None:
        """"""
        self.__input_file = input_file
        self.__args = args
        self.__verbose = verbose
        self.__repeatable = repeatable

    @property
    def name(self) -> str:
        """"""
        return "Auditor"

    @property
    def description(self) -> str:
        """"""
        return f"Running pip-audit on exported file: {self.__input_file}"

    @property
    def input_file(self) -> Path:
        """"""
        return self.__input_file

    @property
    def args(self) -> tuple[str, ...]:
        """"""
        return self.__args

    @traced_function
    def execute(self) -> int:
        """"""
        pip_audit: Path = self._which("pip-audit")
        if pip_audit is None:
            return -1

        arguments = [a for a in self.args]
        if self.__repeatable:
            arguments.append("--require-hashes")
        arguments.append("--disable-pip")
        arguments.append("--skip-editable")
        arguments.append("--progress-spinner")
        arguments.append("off")
        arguments.append("--format")
        arguments.append("json")
        arguments.append("--requirement")
        arguments.append(str(self.input_file))

        arg_items = tuple(arguments)

        exit_code, stdout, stderr = self.run(pip_audit, arg_items)

        exit_code_no_vulnerabilities = 0
        exit_code_has_vulnerabilities = 1

        if exit_code in (
            exit_code_no_vulnerabilities,
            exit_code_has_vulnerabilities,
        ):
            if len(stdout) > 0:
                d = get_dependencies(stdout)
                if d is not None:
                    num_vulnerabilities = self._get_number_of_vulnerabilities(
                        d, self.__verbose
                    )
                    logger.warning(
                        "%i vulnerabilities found", num_vulnerabilities
                    )
                else:
                    logger.warning(
                        "Failed to get dependencies with vulnerabilities"
                    )
            elif exit_code == exit_code_has_vulnerabilities:
                logger.warning(
                    "Vulnerable packages found. Failed to get details"
                )
            else:
                logger.info("0 vulnerabilities found.")

            return 0
        else:
            logger.warning(stderr)

        return exit_code

    def _get_number_of_vulnerabilities(
        self, dependencies, log_vulnerabilities=False
    ):
        def get_vulnerability_id(vuln) -> str:
            if len(vuln.aliases) > 0:
                return f"{vuln.id},{','.join(vuln.aliases)}"
            return vuln.id

        def get_solved_versions(vuln) -> "str | None":
            if len(vuln.fixed_versions) > 0:
                return ",".join(vuln.fixed_versions)
            return None

        num_vulnerabilities = 0
        for v in [d for d in dependencies.dependencies if len(d.vulns) > 0]:
            for vulnerability in v.vulns:
                num_vulnerabilities = num_vulnerabilities + 1
                fixed_versions = get_solved_versions(vulnerability)
                if log_vulnerabilities:

                    logger.info(
                        (
                            "Package %s (Version %s) is vulnerable by %s."
                            "Please upgrade to %s. Details: %s"
                        ),
                        v.name,
                        v.version,
                        get_vulnerability_id(vulnerability),
                        fixed_versions or "UNSOLVED",
                        vulnerability.description,
                    )
                if fixed_versions is not None:
                    logger.warning(
                        "Update %s to version %s", v.name, fixed_versions
                    )
                else:
                    logger.warning(
                        "Packages %s has an unresolved vulnerability", v.name
                    )
        return num_vulnerabilities



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from dataclasses import dataclass, field
from json import loads


@dataclass
class DependenciesInfo:
    """"""

    dependencies: "list[DependencyInfo]" = field(default_factory=list)


@dataclass
class DependencyInfo:
    """"""

    name: str = field()
    version: str = field()
    vulns: "list[VulnerabilityInfo]" = field(default_factory=list)


@dataclass
class VulnerabilityInfo:
    """"""

    id: str = field()
    description: str = field()
    fixed_versions: list[str] = field(default_factory=list)
    aliases: list[str] = field(default_factory=list)


def get_dependencies(json_str: str) -> "DependenciesInfo | None":
    """"""
    data_dict: dict
    try:
        data_dict = loads(json_str)
    except:
        return None
    if "dependencies" in data_dict:
        items = data_dict["dependencies"]
        instances: list[DependencyInfo] = []
        for item in [i for i in items if isinstance(i, dict)]:
            name = item.get("name", None)
            version = item.get("version", None)
            vulns = item.get("vulns", [])
            vulnerabilities: list[VulnerabilityInfo] = []
            if isinstance(vulns, list):
                for vuln in vulns:
                    v_id = vuln.get("id", None)
                    fixed_versions = vuln.get("fix_versions", [])
                    aliases = vuln.get("aliases", [])
                    description = vuln.get("description", None)
                    vulnerabilities.append(
                        VulnerabilityInfo(
                            id=v_id or "UNKNOWN",
                            description=description or "UNKNOWN",
                            aliases=aliases or [],
                            fixed_versions=fixed_versions or [],
                        )
                    )
            instance = DependencyInfo(
                name=name or "UNKNOWN",
                version=version or "UNKNOWN",
                vulns=vulnerabilities,
            )
            instances.append(instance)
        return DependenciesInfo(dependencies=instances)
    return DependenciesInfo(dependencies=[])



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
from pdm.core import Core
from pdm.project import ConfigItem
from pdm.signals import post_install

from pdm_audit.plugin import AuditCommand
from pdm_audit.signal import run_pdm_audit_signal


def main(core: Core) -> None:
    """

    Parameters
    ----------
    core: _CoreLike :


    Returns
    -------

    """
    core.register_command(AuditCommand)
    core.add_config(
        "plugin.audit.post_install_hook",
        ConfigItem(
            "If set to true, run pdm audit after installation",
            True,
            env_var="PDM_AUDIT_PLUGIN_HOOK_PI",
        ),
    )
    core.add_config(
        "plugin.audit.hook_verbose",
        ConfigItem(
            "If set to true, run the hook in verbose mode",
            True,
            env_var="PDM_AUDIT_PLUGIN_HOOK_VERBOSE",
        ),
    )
    core.add_config(
        "plugin.audit.repeatable_audit",
        ConfigItem(
            "If set to true, run pdm audit with repeatable audits. "
            "This will include hashes and hence cannot be used for "
            "references with local references",
            False,
        ),
    )

    post_install.connect(run_pdm_audit_signal)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2025 Carsten Igel.
#
# This file is part of pdm-audit
# (see https://github.com/carstencodes/pdm-audit).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from pdm.project import Project

from .plugin import Auditor


def run_pdm_audit_signal(
    project: Project, candidates: dict, dry_run: bool, **kwargs
) -> None:
    del candidates, kwargs
    run: bool = project.config["plugin.audit.post_install_hook"] or False
    if not dry_run and run:
        auditor: Auditor = Auditor()
        auditor.audit(
            project,
            project.config["plugin.audit.hook_verbose"] or False,
        )
